@@ -1,10 +1,12 @@
 from dataclasses import dataclass
 from pathlib import Path
-from typing import Optional, Union, Tuple, Sequence
+from typing import Optional, Union, Tuple, Sequence, Any, Mapping
 
 import transformers as tr
 
 from transformers_embedder import utils
+from transformers_embedder.modules.scalar_mix import ScalarMix
+from transformers_embedder.modules.encoder import Encoder
 
 if utils.is_torch_available():
     import torch
@@ -33,13 +35,17 @@ class TransformersEmbedder(torch.nn.Module):
     Args:
         model (:obj:`str`, :obj:`tr.PreTrainedModel`):
             Transformer model to use (https://huggingface.co/models).
-        return_words (:obj:`bool`, optional, defaults to :obj:`True`):
-            If ``True`` it returns the word-level embeddings by computing the mean of the
-            sub-words embeddings.
         layer_pooling_strategy (:obj:`str`, optional, defaults to :obj:`last`):
             What output to get from the transformer model. The last hidden state (``last``),
             the concatenation of the selected hidden layers (``concat``), the sum of the selected hidden
-            layers (``sum``), the average of the selected hidden layers (``mean``).
+            layers (``sum``), the average of the selected hidden layers (``mean``), or a scalar mixture of
+            the selected hidden layers (``scalar_mix``).
+        subword_pooling_strategy (:obj:`str`, optional, defaults to :obj:`sparse`):
+            What pooling strategy to use for the sub-word embeddings. Methods available are ``sparse``,
+            ``scatter`` and ``none``. The ``scatter`` strategy is ONNX comptabile but uses ``scatter_add_``
+            that is not deterministic. The ``sparse`` strategy is deterministic but it is not comptabile
+            with ONNX. When ``subword_pooling_strategy`` is ``none``, the sub-word embeddings are not
+            pooled.
         output_layers (:obj:`tuple`, optional, defaults to :obj:`(-4, -3, -2, -1)`):
             Which hidden layers to get from the transformer model.
         fine_tune (:obj:`bool`, optional, defaults to :obj:`True`):
@@ -51,8 +57,8 @@ class TransformersEmbedder(torch.nn.Module):
     def __init__(
         self,
         model: Union[str, tr.PreTrainedModel],
-        return_words: bool = True,
         layer_pooling_strategy: str = "last",
+        subword_pooling_strategy: str = "scatter",
         output_layers: Sequence[int] = (-4, -3, -2, -1),
         fine_tune: bool = True,
         return_all: bool = False,
@@ -62,21 +68,42 @@ class TransformersEmbedder(torch.nn.Module):
         super().__init__()
         if isinstance(model, str):
             config = tr.AutoConfig.from_pretrained(
-                model, output_hidden_states=True, output_attentions=True, *args, **kwargs
+                model,
+                output_hidden_states=True,
+                output_attentions=True,
+                *args,
+                **kwargs,
+            )
+            self.transformer_model = tr.AutoModel.from_pretrained(
+                model, config=config, *args, **kwargs
             )
-            self.transformer_model = tr.AutoModel.from_pretrained(model, config=config, *args, **kwargs)
         else:
             self.transformer_model = model
-        self.return_words = return_words
-        self.pooling_strategy = layer_pooling_strategy
-        if max(map(abs, output_layers)) >= self.transformer_model.config.num_hidden_layers:
+
+        # pooling strategy parameters
+        self.layer_pooling_strategy = layer_pooling_strategy
+        self.subword_pooling_strategy = subword_pooling_strategy
+
+        self._scalar_mix: Optional[ScalarMix] = None
+        if layer_pooling_strategy == "scalar_mix":
+            self._scalar_mix = ScalarMix(len(output_layers))
+
+        # check output_layers is well defined
+        if (
+            max(map(abs, output_layers))
+            >= self.transformer_model.config.num_hidden_layers
+        ):
             raise ValueError(
                 f"`output_layers` parameter not valid, choose between 0 and "
                 f"{self.transformer_model.config.num_hidden_layers - 1}. "
                 f"Current value is `{output_layers}`"
             )
         self.output_layers = output_layers
+
+        # check if return all transformer outputs
         self.return_all = return_all
+
+        # if fine_tune is False, freeze all the transformer's parameters
         if not fine_tune:
             for param in self.transformer_model.parameters():
                 param.requires_grad = False
@@ -86,8 +113,8 @@ class TransformersEmbedder(torch.nn.Module):
         input_ids: torch.Tensor,
         attention_mask: Optional[torch.Tensor] = None,
         token_type_ids: Optional[torch.Tensor] = None,
-        offsets: Optional[torch.Tensor] = None,
-        *args,
+        scatter_offsets: Optional[torch.Tensor] = None,
+        sparse_offsets: Optional[Mapping[str, Any]] = None,
         **kwargs,
     ) -> TransformersEmbedderOutput:
         """
@@ -100,8 +127,12 @@ class TransformersEmbedder(torch.nn.Module):
                 Attention mask for the transformer model.
             token_type_ids (:obj:`torch.Tensor`, optional):
                 Token type ids for the transformer model.
-            offsets (:obj:`torch.Tensor`, optional):
-                Offsets of the sub-word, used to reconstruct the word embeddings.
+            scatter_offsets (:obj:`torch.Tensor`, optional):
+                Offsets of the sub-word, used to reconstruct the word embeddings using
+                the ``scatter`` method.
+            sparse_offsets (:obj:`Mapping[str, Any]`, optional):
+                Offsets of the sub-word, used to reconstruct the word embeddings using
+                the ``sparse`` method.
 
         Returns:
              :obj:`TransformersEmbedderOutput`:
@@ -115,33 +146,75 @@ class TransformersEmbedder(torch.nn.Module):
 
         # Shape: [batch_size, num_sub-words, embedding_size].
         transformer_outputs = self.transformer_model(**inputs)
-        if self.pooling_strategy == "last":
+        if self.layer_pooling_strategy == "last":
             word_embeddings = transformer_outputs.last_hidden_state
-        elif self.pooling_strategy == "concat":
-            word_embeddings = [transformer_outputs.hidden_states[layer] for layer in self.output_layers]
+        elif self.layer_pooling_strategy == "concat":
+            word_embeddings = [
+                transformer_outputs.hidden_states[layer] for layer in self.output_layers
+            ]
             word_embeddings = torch.cat(word_embeddings, dim=-1)
-        elif self.pooling_strategy == "sum":
-            word_embeddings = [transformer_outputs.hidden_states[layer] for layer in self.output_layers]
+        elif self.layer_pooling_strategy == "sum":
+            word_embeddings = [
+                transformer_outputs.hidden_states[layer] for layer in self.output_layers
+            ]
             word_embeddings = torch.stack(word_embeddings, dim=0).sum(dim=0)
-        elif self.pooling_strategy == "mean":
-            word_embeddings = [transformer_outputs.hidden_states[layer] for layer in self.output_layers]
-            word_embeddings = torch.stack(word_embeddings, dim=0).mean(dim=0, dtype=torch.float)
+        elif self.layer_pooling_strategy == "mean":
+            word_embeddings = [
+                transformer_outputs.hidden_states[layer] for layer in self.output_layers
+            ]
+            word_embeddings = torch.stack(word_embeddings, dim=0).mean(
+                dim=0, dtype=torch.float
+            )
+        elif self.layer_pooling_strategy == "scalar_mix":
+            word_embeddings = [
+                transformer_outputs.hidden_states[layer] for layer in self.output_layers
+            ]
+            word_embeddings = self._scalar_mix(word_embeddings)
         else:
             raise ValueError(
-                "`pooling_strategy` parameter not valid, choose between `last`, `concat`, "
-                f"`sum` and `mean`. Current value `{self.pooling_strategy}`"
+                "`layer_pooling_strategy` parameter not valid, choose between `last`, `concat`, "
+                f"`sum`, `mean` and `scalar_mix`. Current value `{self.layer_pooling_strategy}`"
             )
 
-        if self.return_words and offsets is None:
+        if (
+            self.subword_pooling_strategy != "none"
+            and scatter_offsets is None
+            and sparse_offsets is None
+        ):
             raise ValueError(
-                "`return_words` is `True` but `offsets` was not passed to the model. "
-                "Cannot compute word embeddings. To solve:\n"
-                "- Set `return_words` to `False` or"
-                "- Pass `offsets` to the model during forward."
+                "`subword_pooling_strategy` is not `none` but neither `scatter_offsets` not `sparse_offsets` "
+                "were passed to the model. Cannot compute word embeddings.\nTo solve:\n"
+                "- Set `subword_pooling_strategy` to `none` or\n"
+                "- Pass `scatter_offsets` to the model during forward or\n"
+                "- Pass `sparse_offsets` to the model during forward."
             )
-        if self.return_words:
-            # Shape: [batch_size, num_words, embedding_size].
-            word_embeddings = self.merge_subword(word_embeddings, offsets)
+
+        if self.subword_pooling_strategy not in ["none", "scatter", "sparse"]:
+            raise ValueError(
+                "`subword_pooling_strategy` parameter not valid, choose between `scatter`, `sparse`"
+                f" and `none`. Current value is `{self.subword_pooling_strategy}`."
+            )
+        if self.subword_pooling_strategy == "scatter":
+            if scatter_offsets is None:
+                raise ValueError(
+                    "`subword_pooling_strategy` is `scatter` but `scatter_offsets` "
+                    "were not passed to the model. Cannot compute word embeddings.\nTo solve:\n"
+                    "- Set `subword_pooling_strategy` to `none` or\n"
+                    "- Pass `scatter_offsets` to the model during forward."
+                )
+            word_embeddings = self.merge_scatter(
+                word_embeddings, indices=scatter_offsets
+            )
+        if self.subword_pooling_strategy == "sparse":
+            if sparse_offsets is None:
+                raise ValueError(
+                    "`subword_pooling_strategy` is `sparse` but `sparse_offsets` "
+                    "were not passed to the model. Cannot compute word embeddings.\nTo solve:\n"
+                    "- Set `subword_pooling_strategy` to `none` or\n"
+                    "- Pass `sparse_offsets` to the model during forward."
+                )
+            word_embeddings = self.merge_sparse(word_embeddings, sparse_offsets)
+
         if self.return_all:
             return TransformersEmbedderOutput(
                 word_embeddings=word_embeddings,
@@ -152,12 +225,13 @@ class TransformersEmbedder(torch.nn.Module):
             )
         return TransformersEmbedderOutput(word_embeddings=word_embeddings)
 
-    def merge_subword(self, embeddings: torch.Tensor, indices: torch.Tensor) -> torch.Tensor:
+    @staticmethod
+    def merge_scatter(embeddings: torch.Tensor, indices: torch.Tensor) -> torch.Tensor:
         """
         Minimal version of ``scatter_mean``, from `pytorch_scatter
         <https://github.com/rusty1s/pytorch_scatter/>`_
-        library, that is compatible for ONNX but works only for our case. It is used to compute word level
-        embeddings from the transformer output.
+        library, that is compatible for ONNX but works only for our case.
+        It is used to compute word level embeddings from the transformer output.
 
         Args:
             embeddings (:obj:`torch.Tensor`):
@@ -168,38 +242,85 @@ class TransformersEmbedder(torch.nn.Module):
         Returns:
             :obj:`torch.Tensor`
         """
+
+        def broadcast(src: torch.Tensor, other: torch.Tensor):
+            """
+            Broadcast ``src`` to match the shape of ``other``.
+
+            Args:
+                src (:obj:`torch.Tensor`):
+                    The tensor to broadcast.
+                other (:obj:`torch.Tensor`):
+                    The tensor to match the shape of.
+
+            Returns:
+                :obj:`torch.Tensor`: The broadcasted tensor.
+            """
+            for _ in range(src.dim(), other.dim()):
+                src = src.unsqueeze(-1)
+            src = src.expand_as(other)
+            return src
+
+        def scatter_sum(src: torch.Tensor, index: torch.Tensor) -> torch.Tensor:
+            """
+            Sums the elements in ``src`` that have the same indices as in ``index``.
+
+            Args:
+                src (:obj:`torch.Tensor`):
+                    The tensor to sum.
+                index (:obj:`torch.Tensor`):
+                    The indices to sum.
+
+            Returns:
+                :obj:`torch.Tensor`: The summed tensor.
+            """
+            index = broadcast(index, src)
+            size = list(src.size())
+            size[1] = index.max() + 1
+            out = torch.zeros(size, dtype=src.dtype, device=src.device)
+            return out.scatter_add_(1, index, src)
+
         # replace padding indices with the maximum value inside the batch
         indices[indices == -1] = torch.max(indices)
-        out = self.scatter_sum(embeddings, indices)
-        ones = torch.ones(indices.size(), dtype=embeddings.dtype, device=embeddings.device)
-        count = self.scatter_sum(ones, indices)
+        merged = scatter_sum(embeddings, indices)
+        ones = torch.ones(
+            indices.size(), dtype=embeddings.dtype, device=embeddings.device
+        )
+        count = scatter_sum(ones, indices)
         count.clamp_(1)
-        count = self.broadcast(count, out)
-        out.true_divide_(count)
-        return out
+        count = broadcast(count, merged)
+        merged.true_divide_(count)
+        return merged
 
-    def scatter_sum(self, src: torch.Tensor, index: torch.Tensor) -> torch.Tensor:
+    @staticmethod
+    def merge_sparse(
+        embeddings: torch.Tensor, bpe_info: Optional[Mapping[str, Any]]
+    ) -> torch.Tensor:
         """
-        Minimal version of ``scatter_sum``, from `pytorch_scatter
-        <https://github.com/rusty1s/pytorch_scatter/>`_
-        library, that is compatible for ONNX but works only for our case.
+        Merges the subword embeddings into a single tensor, using sparse indices.
 
         Args:
-            src (:obj:`torch.Tensor`):
-                The source tensor.
-            index (:obj:`torch.Tensor`):
-                The indices of elements to scatter.
+            embeddings (:obj:`torch.Tensor`):
+                The embeddings tensor.
+            bpe_info (:obj:`Mapping[str, Any]`, `optional`):
+                The BPE info.
 
         Returns:
-            :obj:`torch.Tensor`
+            :obj:`torch.Tensor`: The merged embeddings.
         """
-        index = self.broadcast(index, src)
-        size = list(src.size())
-        size[1] = index.max() + 1
-        out = torch.zeros(size, dtype=src.dtype, device=src.device)
-        return out.scatter_add_(1, index, src)
+        # it is constructed here and not in the tokenizer/collate because pin_memory is not sparse-compatible
+        bpe_weights = torch.sparse_coo_tensor(
+            indices=bpe_info["sparse_indices"],
+            values=bpe_info["sparse_values"],
+            size=bpe_info["sparse_size"],
+        )
+        # (sentence, word, bpe) x (sentence, bpe, transformer_dim) -> (sentence, word, transformer_dim)
+        merged = torch.bmm(bpe_weights.to_dense(), embeddings)
+        return merged
 
-    def resize_token_embeddings(self, new_num_tokens: Optional[int] = None) -> torch.nn.Embedding:
+    def resize_token_embeddings(
+        self, new_num_tokens: Optional[int] = None
+    ) -> torch.nn.Embedding:
         """
         Resizes input token embeddings' matrix of the model if :obj:`new_num_tokens != config.vocab_size`.
 
@@ -222,27 +343,6 @@ class TransformersEmbedder(torch.nn.Module):
         """
         self.transformer_model.save_pretrained(save_directory)
 
-    @staticmethod
-    def broadcast(src: torch.Tensor, other: torch.Tensor):
-        """
-        Minimal version of ``broadcast``, from `pytorch_scatter
-        <https://github.com/rusty1s/pytorch_scatter/>`_
-        library, that is compatible with ONNX but works only for our case.
-
-        Args:
-            src (:obj:`torch.Tensor`):
-                The source tensor.
-            other (:obj:`torch.Tensor`):
-                The tensor from which we want to broadcast.
-
-        Returns:
-            :obj:`torch.Tensor`
-        """
-        for _ in range(src.dim(), other.dim()):
-            src = src.unsqueeze(-1)
-        src = src.expand_as(other)
-        return src
-
     @property
     def hidden_size(self) -> int:
         """
@@ -251,7 +351,9 @@ class TransformersEmbedder(torch.nn.Module):
         Returns:
             :obj:`int`: Hidden size of ``self.transformer_model``.
         """
-        multiplier = len(self.output_layers) if self.pooling_strategy == "concat" else 1
+        multiplier = (
+            len(self.output_layers) if self.layer_pooling_strategy == "concat" else 1
+        )
         return self.transformer_model.config.hidden_size * multiplier
 
     @property
@@ -262,60 +364,12 @@ class TransformersEmbedder(torch.nn.Module):
         Returns:
             :obj:`int`: Hidden size of ``self.transformer_model``.
         """
-        multiplier = len(self.output_layers) if self.pooling_strategy == "concat" else 1
+        multiplier = (
+            len(self.output_layers) if self.layer_pooling_strategy == "concat" else 1
+        )
         return self.transformer_model.config.hidden_size * multiplier
 
 
-class Encoder(torch.nn.Module):
-    """
-    An encoder module for the :obj:`TransformersEmbedder` class.
-
-    Args:
-        transformer_hidden_size (:obj:`int`):
-            The hidden size of the inner transformer.
-        projection_size (:obj:`int`, `optional`, defaults to :obj:`None`):
-            The size of the projection layer.
-        dropout (:obj:`float`, `optional`, defaults to :obj:`0.1`):
-            The dropout value.
-        bias (:obj:`bool`, `optional`, defaults to :obj:`True`):
-            Whether to use a bias.
-    """
-
-    def __init__(
-        self,
-        transformer_hidden_size: int,
-        projection_size: Optional[int] = None,
-        dropout: float = 0.1,
-        bias: bool = True,
-    ):
-        super().__init__()
-        self.normalization_layer = torch.nn.BatchNorm1d(transformer_hidden_size)
-        self.projection_size = projection_size or transformer_hidden_size
-        self.projection_layer = torch.nn.Linear(transformer_hidden_size, self.projection_size, bias=bias)
-        self.dropout_layer = torch.nn.Dropout(dropout)
-        self.activation_layer = torch.nn.SiLU()
-
-    def forward(self, x: torch.Tensor) -> torch.Tensor:
-        """
-        Forward pass of the encoder.
-
-        Args:
-            x (:obj:`torch.Tensor`):
-                The input tensor.
-
-        Returns:
-            :obj:`torch.Tensor`: The encoded tensor.
-        """
-        x = self.dropout_layer(x)
-        x = x.permute(0, 2, 1)
-        x = self.normalization_layer(x)
-        x = x.permute(0, 2, 1)
-        x = self.projection_layer(x)
-        x = self.activation_layer(x)
-        x = self.dropout_layer(x)
-        return x
-
-
 class TransformersEncoder(TransformersEmbedder):
     """
     Transformer Embedder class.
@@ -325,13 +379,15 @@ class TransformersEncoder(TransformersEmbedder):
     Args:
         model (:obj:`str`, :obj:`tr.PreTrainedModel`):
             Transformer model to use (https://huggingface.co/models).
-        return_words (:obj:`bool`, optional, defaults to :obj:`True`):
-            If ``True`` it returns the word-level embeddings by computing the mean of the
-            sub-words embeddings.
         layer_pooling_strategy (:obj:`str`, optional, defaults to :obj:`last`):
             What output to get from the transformer model. The last hidden state (``last``),
             the concatenation of the selected hidden layers (``concat``), the sum of the selected hidden
             layers (``sum``), the average of the selected hidden layers (``mean``).
+        subword_pooling_strategy (:obj:`str`, optional, defaults to :obj:`scatter`):
+            What pooling strategy to use for the sub-word embeddings. Methods available are ``scatter``,
+            ``sparse`` and ``none``. The ``scatter`` strategy is ONNX comptabile but uses ``scatter_add``
+            that is not deterministic. The ``sparse`` strategy is deterministic but it is not comptabile
+            with ONNX.
         output_layers (:obj:`tuple`, optional, defaults to :obj:`(-4, -3, -2, -1)`):
             Which hidden layers to get from the transformer model.
         fine_tune (:obj:`bool`, optional, defaults to :obj:`True`):
@@ -340,6 +396,8 @@ class TransformersEncoder(TransformersEmbedder):
             If ``True``, returns all the outputs from the HuggingFace model.
         projection_size (:obj:`int`, optional, defaults to :obj:`None`):
             If not ``None``, the output of the transformer is projected to this size.
+        activation_layer (:obj:`torch.nn.Module`, optional, defaults to :obj:`None`):
+            Activation layer to use. If ``None``, no activation layer is used.
         dropout (:obj:`float`, optional, defaults to :obj:`0.1`):
             The dropout probability.
         bias (:obj:`bool`, optional, defaults to :obj:`True`):
@@ -349,25 +407,43 @@ class TransformersEncoder(TransformersEmbedder):
     def __init__(
         self,
         model: Union[str, tr.PreTrainedModel],
-        return_words: bool = True,
         layer_pooling_strategy: str = "last",
+        subword_pooling_strategy: str = "sparse",
         output_layers: Sequence[int] = (-4, -3, -2, -1),
         fine_tune: bool = True,
         return_all: bool = False,
         projection_size: Optional[int] = None,
+        activation_layer: Optional[torch.nn.Module] = None,
         dropout: float = 0.1,
         bias: bool = True,
+        *args,
+        **kwargs,
     ) -> None:
-        super().__init__(model, return_words, layer_pooling_strategy, output_layers, fine_tune, return_all)
-        self.encoder = Encoder(self.transformer_hidden_size, projection_size, dropout, bias)
+        super().__init__(
+            model,
+            layer_pooling_strategy,
+            subword_pooling_strategy,
+            output_layers,
+            fine_tune,
+            return_all,
+            *args,
+            **kwargs,
+        )
+        self.encoder = Encoder(
+            self.transformer_hidden_size,
+            projection_size,
+            activation_layer,
+            dropout,
+            bias,
+        )
 
     def forward(
         self,
         input_ids: torch.Tensor,
         attention_mask: Optional[torch.Tensor] = None,
         token_type_ids: Optional[torch.Tensor] = None,
-        offsets: Optional[torch.Tensor] = None,
-        *args,
+        scatter_offsets: Optional[torch.Tensor] = None,
+        sparse_offsets: Optional[Mapping[str, Any]] = None,
         **kwargs,
     ) -> TransformersEmbedderOutput:
         """
@@ -380,16 +456,22 @@ class TransformersEncoder(TransformersEmbedder):
                 Attention mask for the transformer model.
             token_type_ids (:obj:`torch.Tensor`, optional):
                 Token type ids for the transformer model.
-            offsets (:obj:`torch.Tensor`, optional):
+            scatter_offsets (:obj:`torch.Tensor`, optional):
                 Offsets of the sub-word, used to reconstruct the word embeddings.
 
         Returns:
              :obj:`TransformersEmbedderOutput`:
                 Word level embeddings plus the output of the transformer model.
         """
-        transformer_output = super().forward(
-            input_ids, attention_mask, token_type_ids, offsets, *args, **kwargs
-        )
+        transformers_kwargs = {
+            "input_ids": input_ids,
+            "attention_mask": attention_mask,
+            "token_type_ids": token_type_ids,
+            "scatter_offsets": scatter_offsets,
+            "sparse_offsets": sparse_offsets,
+            **kwargs,
+        }
+        transformer_output = super().forward(**transformers_kwargs)
         encoder_output = self.encoder(transformer_output.word_embeddings)
         transformer_output.word_embeddings = encoder_output
         return transformer_output
