@@ -2,7 +2,7 @@ from __future__ import annotations
 
 from collections import UserDict
 from functools import partial
-from typing import List, Dict, Union, Any, Optional, Tuple, Set
+from typing import List, Dict, Union, Any, Optional, Tuple, Set, Sequence, Mapping
 
 import transformers as tr
 from transformers import BatchEncoding
@@ -27,12 +27,24 @@ class Tokenizer:
     Args:
         model (:obj:`str`, :obj:`transformers.PreTrainedTokenizer`):
             Language model name (or a transformer :obj:`PreTrainedTokenizer`.
+        return_sparse_offsets (:obj:`bool`, optional, defaults to :obj:`True`):
+            If :obj:`True`, the sparse offsets of the tokens in the input text are returned. To reduce
+            memory usage, set this to :obj:`False` if you don't need them, e.g. you set the
+            `subword_pooling_strategy` to `scatter` in the `TransformersEmbedder` model.
     """
 
-    def __init__(self, model: Union[str, tr.PreTrainedTokenizer], *args, **kwargs):
+    def __init__(
+        self,
+        model: Union[str, tr.PreTrainedTokenizer],
+        return_sparse_offsets: bool = True,
+        *args,
+        **kwargs,
+    ):
         if isinstance(model, str):
             # init HuggingFace tokenizer
-            self.huggingface_tokenizer = tr.AutoTokenizer.from_pretrained(model, *args, **kwargs)
+            self.huggingface_tokenizer = tr.AutoTokenizer.from_pretrained(
+                model, *args, **kwargs
+            )
             # get config
             self.config = tr.AutoConfig.from_pretrained(model, *args, **kwargs)
         else:
@@ -40,6 +52,9 @@ class Tokenizer:
             self.config = tr.AutoConfig.from_pretrained(
                 self.huggingface_tokenizer.name_or_path, *args, **kwargs
             )
+
+        self.return_sparse_offsets = return_sparse_offsets
+
         # padding stuff
         # default, batch length is model max length
         self.subword_max_batch_len = self.huggingface_tokenizer.model_max_length
@@ -90,11 +105,29 @@ class Tokenizer:
             :obj:`ModelInputs`: The inputs to the transformer model.
         """
         # some checks before starting
+        if return_tensors == "tf":
+            raise ValueError(
+                "`return_tensors='tf'` is not supported. Please use `return_tensors='pt'` "
+                "or `return_tensors=True`."
+            )
         if return_tensors is True:
             return_tensors = "pt"
         if return_tensors is False:
             return_tensors = None
 
+        # check if input is batched or a single sample
+        is_batched = bool(
+            isinstance(text, (list, tuple))
+            and text
+            and (
+                (isinstance(text[0], (list, tuple)) and is_split_into_words)
+                or isinstance(text[0], str)
+            )
+        )
+        if not is_batched:  # batch it
+            text = [text]
+            text_pair = [text_pair] if text_pair is not None else None
+
         # use huggingface tokenizer to encode the text
         model_inputs = self.huggingface_tokenizer(
             text,
@@ -108,14 +141,28 @@ class Tokenizer:
             **kwargs,
         )
         # build the offsets used to pool the subwords
-        offsets, sentence_lengths = self.build_offsets(
-            model_inputs, return_tensors=return_tensors, there_is_text_pair=text_pair is not None
+        scatter_offsets, sentence_lengths = self.build_scatter_offsets(
+            model_inputs,
+            return_tensors=return_tensors,
+            there_is_text_pair=text_pair is not None,
         )
 
         # convert to ModelInputs
         model_inputs = ModelInputs(**model_inputs)
         # add the offsets to the model inputs
-        model_inputs.update({"offsets": offsets, "sentence_lengths": sentence_lengths})
+        model_inputs.update(
+            {"scatter_offsets": scatter_offsets, "sentence_lengths": sentence_lengths}
+        )
+
+        if self.return_sparse_offsets:
+            # build the data used to pool the subwords when in sparse mode
+            bpe_info: Mapping[str, Any] = self.build_sparse_offsets(
+                offsets=scatter_offsets,
+                bpe_mask=model_inputs.attention_mask,
+                words_per_sentence=sentence_lengths,
+            )
+            # add the bpe info to the model inputs
+            model_inputs["sparse_offsets"] = ModelInputs(**bpe_info)
 
         # we also update the maximum batch length,
         # both for subword and word level
@@ -127,10 +174,12 @@ class Tokenizer:
             model_inputs.update(additional_inputs)
             # check if there is a padding strategy
             if padding:
-                missing_keys = set(additional_inputs.keys()) - set(self.padding_ops.keys())
+                missing_keys = set(additional_inputs.keys()) - set(
+                    self.padding_ops.keys()
+                )
                 if missing_keys:
                     raise ValueError(
-                        f"There are no padding strategy for the following keys: {missing_keys}. "
+                        f"There are no padding strategies for the following keys: {missing_keys}. "
                         "Please add one with `tokenizer.add_padding_ops()`."
                     )
                 self.pad_batch(model_inputs)
@@ -140,7 +189,7 @@ class Tokenizer:
 
         return model_inputs
 
-    def build_offsets(
+    def build_scatter_offsets(
         self,
         model_inputs: BatchEncoding,
         return_tensors: bool = True,
@@ -172,7 +221,9 @@ class Tokenizer:
             # with our subword pooling strategy
             # if the first token is a special token, we need to take it into account
             if self.has_starting_token:
-                word_offsets = [0] + [w + 1 if w is not None else w for w in word_ids[1:]]
+                word_offsets = [0] + [
+                    w + 1 if w is not None else w for w in word_ids[1:]
+                ]
             # otherwise, we can just use word_ids as is
             else:
                 word_offsets = word_ids
@@ -191,7 +242,8 @@ class Tokenizer:
                     word_offsets[sep_index] = sep_offset_value
                 # keep the first offsets as is, adjust the second ones
                 word_offsets = word_offsets[: sep_index + 1] + [
-                    w + sep_offset_value if w is not None else w for w in word_offsets[sep_index + 1 :]
+                    w + sep_offset_value if w is not None else w
+                    for w in word_offsets[sep_index + 1 :]
                 ]
                 # update again the sep_offset
                 sep_offset_value = max([w for w in word_offsets if w is not None]) + 1
@@ -209,8 +261,78 @@ class Tokenizer:
             offsets = torch.as_tensor(offsets)
         return offsets, sentence_lengths
 
+    @staticmethod
+    def build_sparse_offsets(
+        offsets: torch.Tensor | Sequence[Sequence[int]],
+        bpe_mask: torch.Tensor | Sequence[Sequence[int]],
+        words_per_sentence: Sequence[int],
+    ) -> Mapping[str, Any]:
+        """Build tensors used as info for BPE pooling, starting from the BPE offsets.
+
+        Args:
+            offsets (:obj:`torch.Tensor` or :obj:`List[List[int]]`):
+                The offsets to compute lengths from.
+            bpe_mask (:obj:`torch.Tensor` or :obj:`List[List[int]]`):
+                The attention mask at BPE level.
+            words_per_sentence (:obj:`List[int]`):
+                The sentence lengths, word-wise.
+
+        Returns:
+            :obj:`Mapping[str, Any]`: Tensors used to construct the sparse one which pools the
+            transformer encoding word-wise.
+        """
+        if not isinstance(offsets, torch.Tensor):
+            offsets: torch.Tensor = torch.as_tensor(offsets)
+        if not isinstance(bpe_mask, torch.Tensor):
+            bpe_mask: torch.Tensor = torch.as_tensor(bpe_mask)
+
+        sentence_lengths: torch.Tensor = bpe_mask.sum(dim=1)
+
+        # We want to build triplets as coordinates (document, word, bpe)
+        # We start by creating the document index for each triplet
+        document_indices = torch.arange(offsets.size(0)).repeat_interleave(
+            sentence_lengths
+        )
+        # then the word indices
+        word_indices = offsets[offsets != -1]
+        # lastly the bpe indices
+        max_range: torch.Tensor = torch.arange(bpe_mask.shape[1])
+        bpe_indices: torch.LongTensor = torch.cat(
+            [max_range[:i] for i in bpe_mask.sum(dim=1)], dim=0
+        ).long()
+
+        unique_words, word_lengths = torch.unique_consecutive(
+            offsets, return_counts=True
+        )
+        unpadded_word_lengths = word_lengths[unique_words != -1]
+
+        # and their weight to be used as multiplication factors
+        bpe_weights: torch.FloatTensor = (
+            (1 / unpadded_word_lengths).repeat_interleave(unpadded_word_lengths).float()
+        )
+
+        sparse_indices = torch.stack(
+            [document_indices, word_indices, bpe_indices], dim=0
+        )
+
+        bpe_shape = torch.Size(
+            (
+                bpe_mask.size(0),  # batch_size
+                max(words_per_sentence),  # max number of words per sentence
+                bpe_mask.size(1),  # max bpe_number in batch wrt the sentence
+            )
+        )
+
+        return dict(
+            sparse_indices=sparse_indices,
+            sparse_values=bpe_weights,
+            sparse_size=bpe_shape,
+        )
+
     def pad_batch(
-        self, batch: Union[ModelInputs, Dict[str, list]], max_length: Optional[int] = None
+        self,
+        batch: Union[ModelInputs, Dict[str, list]],
+        max_length: Optional[int] = None,
     ) -> ModelInputs:
         """
         Pad the batch to its maximum length or to the specified :obj:`max_length`.
@@ -235,6 +357,7 @@ class Tokenizer:
         for key in batch:
             if key in self.padding_ops:
                 batch[key] = [self.padding_ops[key](b) for b in batch[key]]
+
         return ModelInputs(batch)
 
     def pad_sequence(
@@ -272,7 +395,9 @@ class Tokenizer:
         padding = [value] * abs(length - len(sequence))
         if isinstance(sequence, torch.Tensor):
             if len(sequence.shape) > 1:
-                raise ValueError(f"Sequence tensor must be 1D. Current shape is `{len(sequence.shape)}`")
+                raise ValueError(
+                    f"Sequence tensor must be 1D. Current shape is `{len(sequence.shape)}`"
+                )
             padding = torch.as_tensor(padding)
         if pad_to_left:
             if isinstance(sequence, torch.Tensor):
@@ -282,7 +407,9 @@ class Tokenizer:
             return torch.cat((sequence, padding), -1)
         return sequence + padding
 
-    def add_special_tokens(self, special_tokens_dict: Dict[str, Union[str, tr.AddedToken]]) -> int:
+    def add_special_tokens(
+        self, special_tokens_dict: Dict[str, Union[str, tr.AddedToken]]
+    ) -> int:
         """
         Add a dictionary of special tokens (eos, pad, cls, etc.) to the encoder.
         If special tokens are NOT in the vocabulary, they are added to it (indexed starting from the last
@@ -319,7 +446,7 @@ class Tokenizer:
             self.to_tensor_inputs.add(key)
         self.padding_ops[key] = partial(self.pad_sequence, value=value, length=length)
 
-    def add_to_tensor_inputs(self, names: Union[str, set]) -> Set[str]:
+    def add_to_tensor_inputs(self, names: Union[str, Sequence[str]]) -> Set[str]:
         """
         Add these keys to the ones that will be converted in Tensors.
 
@@ -332,6 +459,8 @@ class Tokenizer:
         """
         if isinstance(names, str):
             names = {names}
+        if not isinstance(names, set):
+            names = set(names)
         self.to_tensor_inputs |= names
         return self.to_tensor_inputs
 
@@ -350,7 +479,9 @@ class Tokenizer:
         """
         # convert to tensor
         batch = {
-            k: torch.as_tensor(v) if k in self.to_tensor_inputs and not isinstance(v, torch.Tensor) else v
+            k: torch.as_tensor(v)
+            if k in self.to_tensor_inputs and not isinstance(v, torch.Tensor)
+            else v
             for k, v in batch.items()
         }
         return ModelInputs(batch)
@@ -425,7 +556,10 @@ class Tokenizer:
                             isinstance(text_to_check[0], str)
                             or (
                                 isinstance(text_to_check[0], (list, tuple))
-                                and (len(text_to_check[0]) == 0 or isinstance(text_to_check[0][0], str))
+                                and (
+                                    len(text_to_check[0]) == 0
+                                    or isinstance(text_to_check[0][0], str)
+                                )
                             )
                         )
                     )
@@ -453,9 +587,9 @@ class Tokenizer:
         Returns:
             :obj:`int`: the number of special tokens.
         """
-        if isinstance(self.huggingface_tokenizer, MODELS_WITH_DOUBLE_SEP) and isinstance(
-            self.huggingface_tokenizer, MODELS_WITH_STARTING_TOKEN
-        ):
+        if isinstance(
+            self.huggingface_tokenizer, MODELS_WITH_DOUBLE_SEP
+        ) and isinstance(self.huggingface_tokenizer, MODELS_WITH_STARTING_TOKEN):
             return 4
         if isinstance(
             self.huggingface_tokenizer,
@@ -592,8 +726,11 @@ class ModelInputs(UserDict):
         """
         if isinstance(device, (str, torch.device, int)):
             self.data = {
-                k: v.to(device=device) if isinstance(v, torch.Tensor) else v for k, v in self.data.items()
+                k: v.to(device=device) if hasattr(v, "to") else v
+                for k, v in self.data.items()
             }
         else:
-            logger.warning(f"Attempting to cast to another type, {str(device)}. This is not supported.")
+            logger.warning(
+                f"Attempting to cast to another type, {str(device)}. This is not supported."
+            )
         return self
